{*************************************************************************
 *                                                                       *
 * DLL linking is provided by Benjamin Rosseaux, www.0ok.de,             *
 * mailto:benjamin@0ok.de                                                *
 *                                                                       *
 * This DLL Loader code is coyyrighted: (C) 2004, Benjamin Rosseaux      *
 *                                                                       *
 *************************************************************************}

unit DLLLoader;

interface

uses Windows, Kol, Share, Debug;

const
  IMPORTED_NAME_OFFSET = $00000002;
  IMAGE_ORDINAL_FLAG32 = $80000000;
  IMAGE_ORDINAL_MASK32 = $0000FFFF;

  RTL_CRITSECT_TYPE  = 0;
  RTL_RESOURCE_TYPE  = 1;

  DLL_PROCESS_ATTACH = 1;
  DLL_THREAD_ATTACH  = 2;
  DLL_THREAD_DETACH  = 3;
  DLL_PROCESS_DETACH = 0;

  IMAGE_SIZEHEADER   = 20;

  IMAGE_FILE_RELOCS_STRIPPED         = $0001;
  IMAGE_FILE_EXECUTABLE_IMAGE        = $0002;
  IMAGE_FILE_LINE_NUMS_STRIPPED      = $0004;
  IMAGE_FILE_LOCAL_SYMS_STRIPPED     = $0008;
  IMAGE_FILE_AGGRESIVE_WS_TRIM       = $0010;
  IMAGE_FILE_byteS_REVERSED_LO       = $0080;
  IMAGE_FILE_32BIT_MACHINE           = $0100;
  IMAGE_FILE_DEBUG_STRIPPED          = $0200;
  IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP = $0400;
  IMAGE_FILE_NET_RUN_FROM_SWAP       = $0800;
  IMAGE_FILE_SYSTEM                  = $1000;
  IMAGE_FILE_DLL                     = $2000;
  IMAGE_FILE_UP_SYSTEM_ONLY          = $4000;
  IMAGE_FILE_byteS_REVERSED_HI       = $8000;

  IMAGE_FILE_MACHINE_UNKNOWN         = $0000;
  IMAGE_FILE_MACHINE_I386            = $014C;
  IMAGE_FILE_MACHINE_R3000           = $0162;
  IMAGE_FILE_MACHINE_R4000           = $0166;
  IMAGE_FILE_MACHINE_R10000          = $0168;
  IMAGE_FILE_MACHINE_ALPHA           = $0184;
  IMAGE_FILE_MACHINE_POWERPC         = $01F0;

  IMAGE_NUMBEROF_DIRECTORY_ENTRIES   = 16;

  IMAGE_SUBSYSTEM_UNKNOWN            = 0;
  IMAGE_SUBSYSTEM_NATIVE             = 1;
  IMAGE_SUBSYSTEM_WINDOWS_GUI        = 2;
  IMAGE_SUBSYSTEM_WINDOWS_CUI        = 3;
  IMAGE_SUBSYSTEM_OS2_CUI            = 5;
  IMAGE_SUBSYSTEM_POSIX_CUI          = 7;
  IMAGE_SUBSYSTEM_RESERVED           = 8;

  IMAGE_DIRECTORY_ENTRY_EXPORT       = 0;
  IMAGE_DIRECTORY_ENTRY_IMPORT       = 1;
  IMAGE_DIRECTORY_ENTRY_RESOURCE     = 2;
  IMAGE_DIRECTORY_ENTRY_EXCEPTION    = 3;
  IMAGE_DIRECTORY_ENTRY_SECURITY     = 4;
  IMAGE_DIRECTORY_ENTRY_BASERELOC    = 5;
  IMAGE_DIRECTORY_ENTRY_DEBUG        = 6;
  IMAGE_DIRECTORY_ENTRY_COPYRIGHT    = 7;
  IMAGE_DIRECTORY_ENTRY_GLOBALPTR    = 8;
  IMAGE_DIRECTORY_ENTRY_TLS          = 9;
  IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG  = 10;
  IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT = 11;
  IMAGE_DIRECTORY_ENTRY_IAT          = 12;

  IMAGE_SIZEOF_SHORT_NAME            = 8;

  IMAGE_SCN_TYIMAGE_REG            = $00000000;
  IMAGE_SCN_TYIMAGE_DSECT          = $00000001;
  IMAGE_SCN_TYIMAGE_NOLOAD         = $00000002;
  IMAGE_SCN_TYIMAGE_GROUP          = $00000004;
  IMAGE_SCN_TYIMAGE_NO_PAD         = $00000008;
  IMAGE_SCN_TYIMAGE_COPY           = $00000010;
  IMAGE_SCN_CNT_CODE               = $00000020;
  IMAGE_SCN_CNT_INITIALIZED_DATA   = $00000040;
  IMAGE_SCN_CNT_UNINITIALIZED_DATA = $00000080;
  IMAGE_SCN_LNK_OTHER              = $00000100;
  IMAGE_SCN_LNK_INFO               = $00000200;
  IMAGE_SCN_TYIMAGE_OVER           = $00000400;
  IMAGE_SCN_LNK_REMOVE             = $00000800;
  IMAGE_SCN_LNK_COMDAT             = $00001000;
  IMAGE_SCN_MEM_PROTECTED          = $00004000;
  IMAGE_SCN_MEM_FARDATA            = $00008000;
  IMAGE_SCN_MEM_SYSHEAP            = $00010000;
  IMAGE_SCN_MEM_PURGEABLE          = $00020000;
  IMAGE_SCN_MEM_16BIT              = $00020000;
  IMAGE_SCN_MEM_LOCKED             = $00040000;
  IMAGE_SCN_MEM_PRELOAD            = $00080000;
  IMAGE_SCN_ALIGN_1byteS           = $00100000;
  IMAGE_SCN_ALIGN_2byteS           = $00200000;
  IMAGE_SCN_ALIGN_4byteS           = $00300000;
  IMAGE_SCN_ALIGN_8byteS           = $00400000;
  IMAGE_SCN_ALIGN_16byteS          = $00500000;
  IMAGE_SCN_ALIGN_32byteS          = $00600000;
  IMAGE_SCN_ALIGN_64byteS          = $00700000;
  IMAGE_SCN_LNK_NRELOC_OVFL        = $01000000;
  IMAGE_SCN_MEM_DISCARDABLE        = $02000000;
  IMAGE_SCN_MEM_NOT_CACHED         = $04000000;
  IMAGE_SCN_MEM_NOT_PAGED          = $08000000;
  IMAGE_SCN_MEM_SHARED             = $10000000;
  IMAGE_SCN_MEM_EXECUTE            = $20000000;
  IMAGE_SCN_MEM_READ               = $40000000;
  IMAGE_SCN_MEM_WRITE              = $80000000;

  IMAGE_REL_BASED_ABSOLUTE       = 0;
  IMAGE_REL_BASED_HIGH           = 1;
  IMAGE_REL_BASED_LOW            = 2;
  IMAGE_REL_BASED_HIGHLOW        = 3;
  IMAGE_REL_BASED_HIGHADJ        = 4;
  IMAGE_REL_BASED_MIPS_JMPADDR   = 5;
  IMAGE_REL_BASED_SECTION        = 6;
  IMAGE_REL_BASED_REL32          = 7;

  IMAGE_REL_BASED_MIPS_JMPADDR16 = 9;
  IMAGE_REL_BASED_IA64_IMM64     = 9;
  IMAGE_REL_BASED_DIR64          = 10;
  IMAGE_REL_BASED_HIGH3ADJ       = 11;

  PAGE_NOACCESS                  = $0001;
  PAGE_READONLY                  = $0002;
  PAGE_READWRITE                 = $0004;
  PAGE_WRITECOPY                 = $0008;
  PAGE_EXECUTE                   = $0010;
  PAGE_EXECUTE_READ              = $0020;
  PAGE_EXECUTE_READWRITE         = $0040;
  PAGE_EXECUTE_WRITECOPY         = $0080;
  PAGE_GUARD                     = $0100;
  PAGE_NOCACHE                   = $0200;

  MEM_COMMIT                     = $00001000;
  MEM_RESERVE                    = $00002000;
  MEM_DECOMMIT                   = $00004000;
  MEM_RELEASE                    = $00008000;
  MEM_FREE                       = $00010000;
  MEM_private                    = $00020000;
  MEM_MAPPED                     = $00040000;
  MEM_RESET                      = $00080000;
  MEM_TOP_DOWN                   = $00100000;
  SEC_FILE                       = $00800000;
  SEC_IMAGE                      = $01000000;
  SEC_RESERVE                    = $04000000;
  SEC_COMMIT                     = $08000000;
  SEC_NOCACHE                    = $10000000;
  MEM_IMAGE                      = SEC_IMAGE;

type

  PPointer = ^Pointer;

  PCardinal  = ^Cardinal;
  PPCardinal = ^PCardinal;

  PWord  = ^word;
  PPWord = ^PWord;

  HINST   = cardinal;
  HMODULE = HINST;

  PWordArray = ^TwordArray;
  TwordArray = Array[0..(2147483647 div SizeOf(word)) - 1] of word;

  PCardinalArray = ^TCardinalArray;
  TCardinalArray = Array[0..(2147483647 div SizeOf(cardinal)) - 1] of cardinal;

  PImageDOSHeader = ^TImageDOSHeader;
  TImageDOSHeader = packed record
    Signature: word;
    PartPag:   word;
    PageCnt:   word;
    ReloCnt:   word;
    HdrSize:   word;
    MinMem:    word;
    MaxMem:    word;
    ReloSS:    word;
    ExeSP:     word;
    ChkSum:    word;
    ExeIP:     word;
    ReloCS:    word;
    Tabloff:   word;
    Overlay:   word;
    Reserved:  packed Array[0..3] of word;
    OEMID:     word;
    OEMInfo:   word;
    Reserved2: packed Array[0..9] of word;
    LFAoffset: cardinal;
  end;

  TISHMisc = packed record
  case integer of
    0: (PhysicalAddress: cardinal);
    1: (VirtualSize: cardinal);
  end;

  PImageExportDirectory = ^TImageExportDirectory;
  TImageExportDirectory = packed record
    Characteristics:       cardinal;
    TimeDateStamp:         cardinal;
    MajorVersion:          word;
    MinorVersion:          word;
    Name:                  cardinal;
    Base:                  cardinal;
    NUMBEROFfunctions:     cardinal;
    NUMBEROFNames:         cardinal;
    Addressoffunctions:    PPCardinal;
    AddressofNames:        PPCardinal;
    AddressofNameOrdinals: PPWord;
  end;

  PImageSectionHeader = ^TImageSectionHeader;
  TImageSectionHeader = packed record
    Name:                 packed Array[0..IMAGE_SIZEOF_SHORT_NAME - 1] of byte;
    Misc:                 TISHMisc;
    VirtualAddress:       cardinal;
    SizeOfRawData:        cardinal;
    PointerToRawData:     cardinal;
    PointerToRelocations: cardinal;
    PointerToLinenumbers: cardinal;
    NUMBEROFRelocations:  word;
    NUMBEROFLinenumbers:  word;
    Characteristics:      cardinal;
  end;

  PImageSectionHeaders = ^TImageSectionHeaders;
  TImageSectionHeaders = Array[0..(2147483647 div SizeOf(TImageSectionHeader)) - 1] of TImageSectionHeader;

  PImageDataDirectory = ^TImageDataDirectory;
  TImageDataDirectory = packed record
    VirtualAddress: cardinal;
    Size:           cardinal;
  end;

  PImageFileHeader = ^TImageFileHeader;
  TImageFileHeader = packed record
    Machine:              word;
    NUMBEROFSections:     word;
    TimeDateStamp:        cardinal;
    PointerToSymbolTable: cardinal;
    NUMBEROFSymbols:      cardinal;
    SizeOfOptionalHeader: word;
    Characteristics:      word;
  end;

  PImageOptionalHeader = ^TImageOptionalHeader;
  TImageOptionalHeader = packed record
    Magic:                       word;
    MajorLinkerVersion:          byte;
    MinorLinkerVersion:          byte;
    SizeOfCode:                  cardinal;
    SizeOfInitializedData:       cardinal;
    SizeOfUninitializedData:     cardinal;
    AddressofEntryPoint:         cardinal;
    BaseofCode:                  cardinal;
    BaseofData:                  cardinal;
    ImageBase:                   cardinal;
    SectionAlignment:            cardinal;
    FileAlignment:               cardinal;
    MajorOperatingSystemVersion: word;
    MinorOperatingSystemVersion: word;
    MajorImageVersion:           word;
    MinorImageVersion:           word;
    MajorSubsystemVersion:       word;
    MinorSubsystemVersion:       word;
    Win32VersionValue:           cardinal;
    SizeOfImage:                 cardinal;
    SizeOfHeaders:               cardinal;
    CheckSum:                    cardinal;
    Subsystem:                   word;
    DllCharacteristics:          word;
    SizeOfStackReserve:          cardinal;
    SizeOfStackCommit:           cardinal;
    SizeOfHeapReserve:           cardinal;
    SizeOfHeapCommit:            cardinal;
    LoaderFlags:                 cardinal;
    NUMBEROFRvaAndSizes:         cardinal;
    DataDirectory:               packed Array[0..IMAGE_NUMBEROF_DIRECTORY_ENTRIES - 1] of TImageDataDirectory;
  end;

  PImageNTHeaders = ^TImageNTHeaders;
  TImageNTHeaders = packed record
    Signature:      cardinal;
    FileHeader:     TImageFileHeader;
    OptionalHeader: TImageOptionalHeader;
  end;

  PImageImportDescriptor = ^TImageImportDescriptor;
  TImageImportDescriptor = packed record
    OriginalFirstThunk: cardinal;
    TimeDateStamp:      cardinal;
    ForwarderChain:     cardinal;
    Name:               cardinal;
    FirstThunk:         cardinal;
  end;

  PImageBaseRelocation = ^TImageBaseRelocation;
  TImageBaseRelocation = packed record
    VirtualAddress: cardinal;
    SizeOfBlock:    cardinal;
  end;

  PImageThunkData = ^TImageThunkData;
  TImageThunkData = packed record
    Forwarderstring: cardinal;
    Funktion:        cardinal;
    Ordinal:         cardinal;
    AddressofData:   cardinal;
  end;

  PSection = ^TSection;
  TSection = packed record
    Base:            Pointer;
    RVA:             cardinal;
    Size:            cardinal;
    Characteristics: cardinal;
  end;

  TSections = Array of TSection;

  TDLLEntryProc = function(hinstDLL: HMODULE; dwReason: cardinal; lpvReserved: Pointer): boolean; stdcall;

  TNameOrID = (niName, niID);

  TExternalLibrary = record
    LibraryName:   string;
    LibraryHandle: HINST;
  end;

  TExternalLibrarys = Array of TExternalLibrary;

  PDLLfunctionImport = ^TDLLfunctionImport;
  TDLLfunctionImport = record
    NameOrID:TNameOrID;
    Name: string;
    ID:   integer;
  end;

  PDLLImport = ^TDLLImport;
  TDLLImport = record
    LibraryName:   string;
    LibraryHandle: HINST;
    Entries:       Array of TDLLfunctionImport;
  end;

  TImports = Array of TDLLImport;

  PDLLfunctionExport = ^TDLLfunctionExport;
  TDLLfunctionExport = record
    Name:            string;
    Index:           integer;
    FunctionPointer: Pointer;
  end;

  TExports = Array of TDLLfunctionExport;

  TExportTreeLink = Pointer;

  PExportTreeNode = ^TExportTreeNode;
  TExportTreeNode = record
    TheChar:                  Char;
    Link:                     TExportTreeLink;
    LinkExist:                boolean;
    Prevoius, Next, Up, Down: PExportTreeNode;
  end;

//==============================================================================

//  THiAsmClass = class(TDebug)
//  private

//  public
//     Stream:THI_Event;
//     procedure DllLoad(var _Data:TData; Index:word);

//  end;

//==============================================================================

  TExportTree = Class
  private
    Root:PExportTreeNode;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Dump;
    function Add(FunctionName: string; Link: TExportTreeLink): boolean;
    function Delete(FunctionName: string): boolean;
    function Find(FunctionName: string; var Link: TExportTreeLink): boolean;
  end;

  TDLLLoader = Class
  private
    ImageBase: Pointer;
    ImageBaseDelta: integer;
    DLLProc: TDLLEntryProc;
    ExternalLibraryArray: TExternalLibrarys;
    ImportArray: TImports;
    ExportArray: TExports;
    Sections: TSections;
    ExportTree: TExportTree;
    function FindExternalLibrary(LibraryName: string): integer;
    function LoadExternalLibrary(LibraryName: string): integer;
    function GetExternalLibraryHandle(LibraryName: string): HINST;
  public
    constructor Create;
    destructor Destroy; override;
    function Load(Stream: PStream): boolean;
    function Unload: boolean;
    function FindExport(FunctionName: string): Pointer;
    function FindExportPerIndex(functionIndex: integer): Pointer;
    function GetExportList: PStrList;
  end;

implementation

function StrToInt(S: string): integer;
var
  C:integer;
begin
  Val(S,Result,C);
end;

function CreateExportTreeNode(AChar: Char): PExportTreeNode;
begin
  GetMem(Result, SizeOf(TExportTreeNode));
  Result^.TheChar := AChar;
  Result^.Link := nil;
  Result^.LinkExist := false;
  Result^.Prevoius := nil;
  Result^.Next := nil;
  Result^.Up := nil;
  Result^.Down := nil;
end;

procedure DestroyExportTreeNode(Node: PExportTreeNode);
begin
  if not Assigned(Node) then exit;
  DestroyExportTreeNode(Node^.Next);
  DestroyExportTreeNode(Node^.Down);
  FreeMem(Node);
end;

//================================= TExportTree ======================================

constructor TExportTree.Create;
begin
  inherited Create;
  Root := nil;
end;

destructor TExportTree.Destroy;
begin
  DestroyExportTreeNode(Root);
  inherited Destroy;
end;

procedure TExportTree.Dump;
var
  Ident:integer;

  procedure DumpNode(Node: PExportTreeNode);
  var
    SubNode: PExportTreeNode;
    IdentCounter,IdentOld: integer;
  begin
    for IdentCounter:=1 to Ident do write(' ');
    write(Node^.TheChar);
    IdentOld := Ident;
    SubNode := Node^.Next;
    while Assigned(SubNode) do
    begin
      write(SubNode.TheChar);
      if not Assigned(SubNode^.Next) then break;
      inc(Ident);
      SubNode:=SubNode^.Next;
    end;
    writeln;
    inc(Ident);
    while Assigned(SubNode) and (SubNode<>Node) do
    begin
      if Assigned(SubNode^.Down) then DumpNode(SubNode^.Down);
      SubNode:=SubNode^.Prevoius;
      dec(Ident);
    end;
    Ident:=IdentOld;
    if Assigned(Node^.Down) then DumpNode(Node^.Down);
  end;
begin
  Ident := 0;
  DumpNode(Root);
end;

function TExportTree.Add(FunctionName: string; Link: TExportTreeLink): boolean;
var
  StringLength, Position, PositionCounter: integer;
  NewNode, LastNode, Node: PExportTreeNode;
  StringChar, NodeChar: Char;
begin
  Result := false;
  StringLength := length(FunctionName);
  if StringLength > 0 then
  begin
    LastNode := nil;
    Node := Root;
    for Position := 1 to StringLength do
    begin
      StringChar := FunctionName[Position];
      if Assigned(Node) then
      begin
        NodeChar := Node^.TheChar;
        if NodeChar = StringChar then
        begin
          LastNode := Node;
          Node := Node^.Next;
        end
        else
        begin
          while (NodeChar < StringChar) and Assigned(Node^.Down) do
          begin
            Node := Node^.Down;
            NodeChar := Node^.TheChar;
          end;
          if NodeChar = StringChar then
          begin
            LastNode := Node;
            Node := Node^.Next;
          end
          else
          begin
            NewNode := CreateExportTreeNode(StringChar);
            if NodeChar < StringChar then
            begin
              NewNode^.Down := Node^.Down;
              NewNode^.Up := Node;
              if Assigned(NewNode^.Down) then NewNode^.Down^.Up := NewNode;
              NewNode^.Prevoius := Node^.Prevoius;
              Node^.Down:=NewNode;
            end
            else if NodeChar > StringChar then
            begin
              NewNode^.Down := Node;
              NewNode^.Up := Node^.Up;
              if Assigned(NewNode^.Up) then NewNode^.Up^.Down := NewNode;
              NewNode^.Prevoius:=Node^.Prevoius;
              if not Assigned(NewNode^.Up) then
              begin
                if Assigned(NewNode^.Prevoius) then
                  NewNode^.Prevoius^.Next:=NewNode
                else
                  Root := NewNode;
              end;
              Node^.Up := NewNode;
            end;
            LastNode := NewNode;
            Node := LastNode^.Next;
          end;
        end;
      end
      else
      begin
        for PositionCounter := Position to StringLength do
        begin
          NewNode := CreateExportTreeNode(FunctionName[PositionCounter]);
          if Assigned(LastNode) then
          begin
            NewNode^.Prevoius := LastNode;
            LastNode^.Next := NewNode;
            LastNode := LastNode^.Next;
          end
          else if not Assigned(Root) then
          begin
            Root := NewNode;
            LastNode := Root;
          end;
        end;
        break;
      end;
    end;
    if Assigned(LastNode) and not LastNode^.LinkExist then
    begin
      LastNode^.Link := Link;
      LastNode^.LinkExist := true;
      Result := true;
    end;
  end;
end;

function TExportTree.Delete(FunctionName: string): boolean;
var
  StringLength, Position: integer;
  Node: PExportTreeNode;
  StringChar, NodeChar: Char;
begin
  Result := false;
  StringLength := length(FunctionName);
  if StringLength > 0 then
  begin
    Node := Root;
    for Position := 1 to StringLength do
    begin
      StringChar := FunctionName[Position];
      if Assigned(Node) then
      begin
        NodeChar := Node^.TheChar;
        while (NodeChar <> StringChar) and Assigned(Node^.Down) do
        begin
          Node := Node^.Down;
          NodeChar := Node^.TheChar;
        end;
        if NodeChar=StringChar then
        begin
          if (Position = StringLength) and Node^.LinkExist then
          begin
            Node^.LinkExist := false;
            Result := true;
            break;
          end;
          Node := Node^.Next;
        end;
      end
      else
        break;
    end;
  end;
end;

function UpChar(C: Char): Char;
begin
  if (C >= 'a') and (C <= 'z') then
    Result := Char(byte(C) - 32)
  else
    Result := C;
end;

function TExportTree.Find(FunctionName: string; var Link: TExportTreeLink): boolean;
var
  StringLength, Position: integer;
  Node: PExportTreeNode;
  StringChar, NodeChar: Char;
begin
  Result := false;
  StringLength := length(FunctionName);
  if StringLength > 0 then
  begin
    Node := Root;
    for Position:=1 to StringLength do
    begin
      StringChar := FunctionName[Position];
      if Assigned(Node) then
      begin
        NodeChar := Node^.TheChar;
        while (UpChar(NodeChar)<>UpChar(StringChar)) and Assigned(Node^.Down) do
        begin
          Node := Node^.Down;
          NodeChar := Node^.TheChar;
        end;
        if UpChar(NodeChar)=UpChar(StringChar) then
        begin
          if (Position=StringLength) and Node^.LinkExist then
          begin
            Link := Node^.Link;
            Result := true;
            break;
          end;
          Node := Node^.Next;
        end;
      end
      else
        break;
    end;
  end;
end;

//================================== TDLLLoader ======================================

constructor TDLLLoader.Create;
begin
  inherited Create;
  ImageBase := nil;
  DLLProc := nil;
  ExternalLibraryArray := nil;
  ImportArray := nil;
  ExportArray := nil;
  Sections := nil;
  ExportTree := nil;
end;

destructor TDLLLoader.Destroy;
begin
  if @DLLProc <> nil then Unload;
  if Assigned(ExportTree) then ExportTree.Destroy;
  inherited Destroy;
end;

function TDLLLoader.FindExternalLibrary(LibraryName: string): integer;
var
 I: integer;
begin
  Result := -1;
  for I := 0 to length(ExternalLibraryArray) - 1 do
  begin
   if ExternalLibraryArray[I].LibraryName = LibraryName then
   begin
     Result:=I;
     exit;
   end;
  end;
end;

function TDLLLoader.LoadExternalLibrary(LibraryName: string): integer;
begin
  Result := FindExternalLibrary(LibraryName);
  if Result < 0 then
  begin
    Result := length(ExternalLibraryArray);
    Setlength(ExternalLibraryArray,length(ExternalLibraryArray) + 1);
    ExternalLibraryArray[Result].LibraryName := LibraryName;
    ExternalLibraryArray[Result].LibraryHandle := LoadLibrary(PChar(LibraryName));
  end;
end;

function TDLLLoader.GetExternalLibraryHandle(LibraryName: string): cardinal;
var
  I: integer;
begin
  Result := 0;
  for I := 0 to length(ExternalLibraryArray) - 1 do
  begin
    if ExternalLibraryArray[I].LibraryName=LibraryName then
    begin
      Result := ExternalLibraryArray[I].LibraryHandle;
      exit;
    end;
  end;
end;

function TDLLLoader.Load(Stream: PStream): boolean;
var
  ImageDOSHeader: TImageDOSHeader;
  ImageNTHeaders: TImageNTHeaders;
  OldProtect: cardinal;

  function ConvertPointer(RVA: cardinal): Pointer;
  var
    I: integer;
  begin
    Result := nil;
    for I := 0 to length(Sections) - 1 do
    begin
      if (RVA<(Sections[I].RVA + Sections[I].Size)) and (RVA >= Sections[I].RVA) then
      begin
        Result := Pointer(cardinal((RVA - cardinal(Sections[I].RVA)) + cardinal(Sections[I].Base)));
        exit;
      end;
    end;
  end;

  function ReadImageHeaders: boolean;
  begin
    Result := false;
    if Stream.Size > 0 then
    begin
      FillChar(ImageNTHeaders,SizeOf(TImageNTHeaders), #0);
      if Stream.Read(ImageDOSHeader, SizeOf(TImageDOSHeader)) <> SizeOf(TImageDOSHeader) then exit;
      if ImageDOSHeader.Signature <> $5A4D then exit;
      if Stream.Seek(ImageDOSHeader.LFAoffset, spBegin) <> cardinal(ImageDOSHeader.LFAoffset) then exit;
      if Stream.Read(ImageNTHeaders.Signature, SizeOf(cardinal)) <> SizeOf(cardinal) then exit;
      if ImageNTHeaders.Signature <> $00004550 then exit;
      if Stream.Read(ImageNTHeaders.FileHeader, SizeOf(TImageFileHeader)) <> SizeOf(TImageFileHeader) then exit;
      if ImageNTHeaders.FileHeader.Machine <> $14C then exit;
      if Stream.Read(ImageNTHeaders.OptionalHeader, ImageNTHeaders.FileHeader.SizeOfOptionalHeader) <> ImageNTHeaders.FileHeader.SizeOfOptionalHeader then exit;
      Result := true;
    end;
  end;

  function InitializeImage: boolean;
  var
    SectionBase: Pointer;
    OldPosition: integer;
  begin
    Result := false;
    if ImageNTHeaders.FileHeader.NUMBEROFSections > 0 then
    begin
      ImageBase:=VirtualAlloc(nil,ImageNTHeaders.OptionalHeader.SizeOfImage, MEM_RESERVE,PAGE_NOACCESS);
      ImageBaseDelta := cardinal(ImageBase) - ImageNTHeaders.OptionalHeader.ImageBase;
      SectionBase := VirtualAlloc(ImageBase, ImageNTHeaders.OptionalHeader.SizeOfHeaders, MEM_COMMIT,PAGE_READWRITE);
      OldPosition := Stream.Position;
      Stream.Seek(0, spBegin);
      Stream.Read(SectionBase^, ImageNTHeaders.OptionalHeader.SizeOfHeaders);
      VirtualProtect(SectionBase, ImageNTHeaders.OptionalHeader.SizeOfHeaders, PAGE_READONLY, @OldProtect);
      Stream.Seek(OldPosition, spBegin);
      Result := true;
    end;
  end;

  function ReadSections: boolean;
  var
    I: integer;
    Section: TImageSectionHeader;
    SectionHeaders: PImageSectionHeaders;
  begin
    Result := false;
    if ImageNTHeaders.FileHeader.NUMBEROFSections > 0 then
    begin
      GetMem(SectionHeaders, ImageNTHeaders.FileHeader. NUMBEROFSections * SizeOf(TImageSectionHeader));
      if Stream.Read(SectionHeaders^, (ImageNTHeaders.FileHeader.NUMBEROFSections * SizeOf(TImageSectionHeader))) <> (ImageNTHeaders.FileHeader.NUMBEROFSections * SizeOf(TImageSectionHeader)) then exit;
      Setlength(Sections, ImageNTHeaders.FileHeader.NUMBEROFSections);
      for I := 0 to ImageNTHeaders.FileHeader.NUMBEROFSections - 1 do
      begin
        Section := SectionHeaders^[I];
        Sections[I].RVA :=Section.VirtualAddress;
        Sections[I].Size := Section.SizeOfRawData;
        if Sections[I].Size<Section.Misc.VirtualSize then Sections[I].Size := Section.Misc.VirtualSize;
        Sections[I].Characteristics := Section.Characteristics;
        Sections[I].Base := VirtualAlloc(Pointer(cardinal(Sections[I].RVA + cardinal(ImageBase))), Sections[I].Size, MEM_COMMIT,PAGE_READWRITE);
        FillChar(Sections[I].Base^, Sections[I].Size, #0);
        if Section.PointerToRawData <> 0 then
        begin
          Stream.Seek(Section.PointerToRawData, spBegin);
          if Stream.Read(Sections[I].Base^, Section.SizeOfRawData) <> cardinal(Section.SizeOfRawData) then exit;
        end;
      end;
      FreeMem(SectionHeaders);
      Result := true;
    end;
  end;

  function ProcessRelocations: boolean;
  var
    Relocations: PChar;
    Position: cardinal;
    BaseRelocation: PImageBaseRelocation;
    Base: Pointer;
    NUMBEROFRelocations: cardinal;
    Relocation: PWordArray;
    RelocationCounter: integer;
    RelocationPointer: Pointer;
    RelocationType: cardinal;
  begin
    if ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress <> 0 then
    begin
      Result := false;
      Relocations := ConvertPointer(ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress);
      Position := 0;
      while Assigned(Relocations) and (Position<ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size) do
      begin
        BaseRelocation := PImageBaseRelocation(Relocations);
        Base := ConvertPointer(BaseRelocation^.VirtualAddress);
        if not Assigned(Base) then exit;
        NUMBEROFRelocations := (BaseRelocation^.SizeOfBlock - SizeOf(TImageBaseRelocation)) div SizeOf(word);
        Relocation := Pointer(cardinal(cardinal(BaseRelocation) + SizeOf(TImageBaseRelocation)));
        for RelocationCounter := 0 to NUMBEROFRelocations - 1 do
        begin
          RelocationPointer := Pointer(cardinal(cardinal(Base) + (Relocation^[RelocationCounter] and $FFF)));
          RelocationType := Relocation^[RelocationCounter] shr 12;
          case RelocationType of
            IMAGE_REL_BASED_ABSOLUTE:
            begin
            end;
            IMAGE_REL_BASED_HIGH:
              PWord(RelocationPointer)^ := (cardinal(((cardinal(PWord(RelocationPointer)^ + cardinal(ImageBase) - ImageNTHeaders.OptionalHeader.ImageBase)))) shr 16) and $FFFF;
            IMAGE_REL_BASED_LOW:
              PWord(RelocationPointer)^ := cardinal(((cardinal(PWord(RelocationPointer)^ + cardinal(ImageBase) - ImageNTHeaders.OptionalHeader.ImageBase)))) and $FFFF;
            IMAGE_REL_BASED_HIGHLOW:
              PPointer(RelocationPointer)^ := Pointer((cardinal(cardinal(PPointer(RelocationPointer)^) + cardinal(ImageBase) - ImageNTHeaders.OptionalHeader.ImageBase)));
            IMAGE_REL_BASED_HIGHADJ:
            begin
            // ???
            end;
            IMAGE_REL_BASED_MIPS_JMPADDR:
            begin
            // Only for MIPS CPUs ;)
            end;
          end;
        end;
        Relocations := Pointer(cardinal(cardinal(Relocations)+BaseRelocation^.SizeOfBlock));
        inc(Position,BaseRelocation^.SizeOfBlock);
      end;
    end;
    Result:=true;
  end;

  function ProcessImports: boolean;
  var
    ImportDescriptor: PImageImportDescriptor;
    ThunkData: PCardinal;
    Name: PChar;
    DLLImport: PDLLImport;
    DLLfunctionImport: PDLLfunctionImport;
    FunctionPointer: Pointer;
  begin
    if ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress <> 0 then
    begin
      ImportDescriptor := ConvertPointer(ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
      if Assigned(ImportDescriptor) then
      begin
        Setlength(ImportArray, 0);
        while ImportDescriptor^.Name <> 0 do
        begin
          Name := ConvertPointer(ImportDescriptor^.Name);
          Setlength(ImportArray,length(ImportArray) + 1);
          LoadExternalLibrary(Name);
          DLLImport := @ImportArray[length(ImportArray) - 1];
          DLLImport^.LibraryName := Name;
          DLLImport^.LibraryHandle := GetExternalLibraryHandle(Name);
          DLLImport^.Entries := nil;
          if ImportDescriptor^.TimeDateStamp = 0 then
            ThunkData := ConvertPointer(ImportDescriptor^.FirstThunk)
          else
            ThunkData := ConvertPointer(ImportDescriptor^.OriginalFirstThunk);
          while ThunkData^ <> 0 do
          begin
            Setlength(DLLImport^.Entries, length(DLLImport^.Entries)+1);
            DLLfunctionImport := @DLLImport^.Entries[length(DLLImport^.Entries) - 1];
            if (ThunkData^ and IMAGE_ORDINAL_FLAG32) <> 0 then
            begin
              DLLfunctionImport^.NameOrID := niID;
              DLLfunctionImport^.ID := ThunkData^ and IMAGE_ORDINAL_MASK32;
              DLLfunctionImport^.Name := '';
              FunctionPointer := GetProcAddress(DLLImport^.LibraryHandle, PChar(ThunkData^ and IMAGE_ORDINAL_MASK32));
            end
            else
            begin
              Name := ConvertPointer(cardinal(ThunkData^) + IMPORTED_NAME_OFFSET);
              DLLfunctionImport^.NameOrID := niName;
              DLLfunctionImport^.ID := 0;
              DLLfunctionImport^.Name := Name;
              FunctionPointer := GetProcAddress(DLLImport^.LibraryHandle,Name);
            end;
            PPointer(Thunkdata)^ := FunctionPointer;
            inc(ThunkData);
          end;
          inc(ImportDescriptor);
        end;
      end;
    end;
    Result:=true;
  end;

  function ProtectSections: boolean;
  var
    I: integer;
    Characteristics: cardinal;
    Flags: cardinal;
  begin
    Result := false;
    if ImageNTHeaders.FileHeader.NUMBEROFSections > 0 then
    begin
      for I := 0 to ImageNTHeaders.FileHeader.NUMBEROFSections - 1 do
      begin
        Characteristics := Sections[I].Characteristics;
        Flags := 0;
        if (Characteristics and IMAGE_SCN_MEM_EXECUTE) <> 0 then
        begin
          if (Characteristics and IMAGE_SCN_MEM_READ) <> 0 then
          begin
            if (Characteristics and IMAGE_SCN_MEM_WRITE) <> 0 then
              Flags := Flags or PAGE_EXECUTE_READWRITE
            else
              Flags := Flags or PAGE_EXECUTE_READ;
          end
          else if (Characteristics and IMAGE_SCN_MEM_WRITE) <> 0 then
            Flags := Flags or PAGE_EXECUTE_WRITECOPY
          else
            Flags:=Flags or PAGE_EXECUTE;
        end
        else if (Characteristics and IMAGE_SCN_MEM_READ) <> 0 then
        begin
          if (Characteristics and IMAGE_SCN_MEM_WRITE) <> 0 then
            Flags := Flags or PAGE_READWRITE
          else
            Flags:=Flags or PAGE_READONLY;
        end
        else if (Characteristics and IMAGE_SCN_MEM_WRITE)<>0 then
          Flags := Flags or PAGE_WRITECOPY
        else
          Flags := Flags or PAGE_NOACCESS;

        if (Characteristics and IMAGE_SCN_MEM_NOT_CACHED) <> 0 then Flags:=Flags or PAGE_NOCACHE;
        VirtualProtect(Sections[I].Base,Sections[I].Size, Flags, @OldProtect);
      end;
      Result:=true;
    end;
  end;

  function InitializeLibrary: boolean;
  begin
    Result := false;
    @DLLProc := ConvertPointer(ImageNTHeaders.OptionalHeader.AddressofEntryPoint);
    if DLLProc(cardinal(ImageBase),DLL_PROCESS_ATTACH,nil) then Result := true;
  end;

  function ProcessExports: boolean;
  var
    I: integer;
    ExportDirectory: PImageExportDirectory;
    ExportDirectorySize: cardinal;
    FunctionNamePointer: Pointer;
    FunctionName: PChar;
    functionIndexPointer: Pointer;
    functionIndex: cardinal;
    FunctionPointer: Pointer;
    ForwarderCharPointer: PChar;
    Forwarderstring: string;
    ForwarderLibrary: string;
    ForwarderLibraryHandle: HINST;

      function ParsestringToNumber(Astring: string): cardinal;
      var
        CharCounter: integer;
      begin
        Result := 0;
        for CharCounter := 0 to length(Astring) - 1 do
        begin
          if Astring[CharCounter] in ['0'..'9'] then
            Result := (Result * 10) + byte(byte(Astring[CharCounter]) - byte('0'))
          else
            exit;
        end;
      end;

  begin
    if ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress <> 0 then
    begin

      ExportTree := TExportTree.Create;
      ExportDirectory := ConvertPointer(ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress);
      if Assigned(ExportDirectory) then
      begin

        ExportDirectorySize:=ImageNTHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size;
        Setlength(ExportArray, ExportDirectory^.NUMBEROFNames);
        for I := 0 to ExportDirectory^.NUMBEROFNames - 1 do
        begin
          FunctionNamePointer := ConvertPointer(cardinal(ExportDirectory^.AddressofNames));
          FunctionNamePointer := ConvertPointer(PCardinalArray(FunctionNamePointer)^[I]);
          FunctionName := FunctionNamePointer;
          functionIndexPointer := ConvertPointer(cardinal(ExportDirectory^.AddressofNameOrdinals));
          functionIndex := PWordArray(functionIndexPointer)^[I];
          FunctionPointer := ConvertPointer(cardinal(ExportDirectory^.Addressoffunctions));
          FunctionPointer := ConvertPointer(PCardinalArray(FunctionPointer)^[functionIndex]);
          ExportArray[I].Name := FunctionName;
          ExportArray[I].Index := functionIndex;
          if (cardinal(ExportDirectory)<cardinal(FunctionPointer)) and
             (cardinal(FunctionPointer)<(cardinal(ExportDirectory)+ExportDirectorySize)) then
          begin
            ForwarderCharPointer := FunctionPointer;
            Forwarderstring := ForwarderCharPointer;
            while ForwarderCharPointer^ <> '.' do inc(ForwarderCharPointer);
            ForwarderLibrary:=COPY(Forwarderstring, 1, POS('.',Forwarderstring) - 1);
            LoadExternalLibrary(ForwarderLibrary);
            ForwarderLibraryHandle := GetExternalLibraryHandle(ForwarderLibrary);
            if ForwarderCharPointer^ = '#' then
            begin
              inc(ForwarderCharPointer);
              Forwarderstring := ForwarderCharPointer;
              ForwarderCharPointer := ConvertPointer(ParsestringToNumber(Forwarderstring));
              Forwarderstring := ForwarderCharPointer;
            end
            else
            begin
              Forwarderstring := ForwarderCharPointer;
              ExportArray[I].FunctionPointer := GetProcAddress(ForwarderLibraryHandle, PChar(Forwarderstring));
            end;
          end
          else
            ExportArray[I].FunctionPointer := FunctionPointer;
          ExportTree.Add(ExportArray[I].Name,ExportArray[I].FunctionPointer);
        end;
      end;
    end;
    Result:=true;
  end;

begin
  Result := false;
  if Assigned(Stream) then
  begin
   Stream.Seek(0, spBegin);
    if Stream.Size>0 then
    begin
      if ReadImageHeaders then
      begin
        if InitializeImage then
        begin
          if ReadSections then
          begin
            if ProcessRelocations then
            begin
              if ProcessImports then
              begin
                if ProtectSections then
                begin
                  if InitializeLibrary then
                  begin
                    if ProcessExports then Result:=true;
                  end;
                end;
              end;
            end;
          end;
        end;
      end;
    end;
  end;
end;

function TDLLLoader.Unload: boolean;
var
  I, J: integer;
begin
  Result := false;
  if @DLLProc <> nil then DLLProc(cardinal(ImageBase), DLL_PROCESS_DETACH, nil);
  for I := 0 to length(Sections) - 1 do
  begin
    if Assigned(Sections[I].Base) then
      VirtualFree(Sections[I].Base, 0, MEM_RELEASE);
  end;
  Setlength(Sections, 0);
  for I := 0 to length(ExternalLibraryArray) - 1 do
  begin
    ExternalLibraryArray[I].LibraryName := '';
    FreeLibrary(ExternalLibraryArray[I].LibraryHandle);
  end;
  Setlength(ExternalLibraryArray, 0);
  for I := 0 to length(ImportArray) - 1 do
  begin
    for J := 0 to length(ImportArray[I].Entries) - 1 do
      ImportArray[I].Entries[J].Name := '';
    Setlength(ImportArray[I].Entries, 0);
  end;
  Setlength(ImportArray, 0);
  for I := 0 to length(ExportArray) - 1 do ExportArray[I].Name := '';
  Setlength(ExportArray, 0);
  VirtualFree(ImageBase, 0, MEM_RELEASE);
  if Assigned(ExportTree) then
  begin
    ExportTree.Destroy;
    ExportTree := nil;
  end;
end;

function TDLLLoader.FindExport(FunctionName: string): Pointer;
var
  I: integer;
begin
  Result := nil;
  if Assigned(ExportTree) then
    ExportTree.Find(FunctionName,Result)
  else
  begin
    for I := 0 to length(ExportArray) - 1 do
    begin
      //if ExportArray[I].Name=FunctionName then begin     lstrcmpi
      if lstrcmpi(pChar(ExportArray[I].Name), pChar(FunctionName)) = 0 then
      begin
        Result := ExportArray[I].FunctionPointer;
        exit;
      end;
    end;
  end;
end;

function TDLLLoader.FindExportPerIndex(functionIndex: integer): Pointer;
var
  I: integer;
begin
  Result := nil;
  for I := 0 to length(ExportArray) - 1 do
  begin
    if ExportArray[I].Index = functionIndex then
    begin
      Result := ExportArray[I].FunctionPointer;
      exit;
    end;
  end;
end;

function TDLLLoader.GetExportList: PStrList;
var
  I: integer;
begin
  Result := NewStrList;
  for I := 0 to length(ExportArray) - 1 do Result.Add(ExportArray[I].Name);
  Result.Sort(false);
end;


end.